import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
from skimage.filters import gaussian
from skimage.transform import rotate, rescale, resize
import glob
from skimage import transform
from scipy.interpolate import griddata
import cv2 as cv
from skimage.util import random_noise
from scipy import interpolate
from skimage.segmentation import flood, flood_fill
from skimage.morphology import dilation, disk
from skimage.transform import PiecewiseAffineTransform, warp


def normalize(h):
    return (h - h.min()) / (h.max() - h.min())


def random_shape(size, radius, knots=6, amplitude=0.1):
    shape = np.zeros((size, size))

    cx, cy = size // 2, size // 2
    t = np.linspace(0, 1, knots)

    move_on = False
    while not move_on:
        x = radius * ((1 - amplitude) * np.sin(2 * np.pi * t) + amplitude * np.random.rand(knots)) + cx
        y = radius * ((1 - amplitude) * np.cos(2 * np.pi * t) + amplitude * np.random.rand(knots)) + cy

        tck, u = interpolate.splprep([x, y], s=0, per=True)
        unew = np.linspace(0, 1, 10000)
        x, y = interpolate.splev(unew, tck)
        x = x.astype(int)
        y = y.astype(int)

        if (x >= 0).all() and (x < size).all() and (y >= 0).all() and (y - size).all():
            move_on = True

    shape[x, y] = 1
    shape = dilation(shape, disk(10))
    shape = flood_fill(shape, (cx, cy), 255)

    return shape > 0


def gaussian_window_mask(size, sigma):
    x = np.arange(size[0])
    y = np.arange(size[1])
    p = np.array([[xx, yy] for xx in x for yy in y])
    c = np.array([len(x) / 2, len(y) / 2])

    g = np.exp(-(p[:, 0] - c[0])**2 / (2 * sigma**2)) * np.exp(-(p[:, 1] - c[1])**2 / (2 * sigma**2))
    g = g.reshape(size)
    return g


def random_nonrigid_transform(img, knots, intensity):
    rows, cols = img.shape[0], img.shape[1]

    src_cols = np.linspace(0, cols, knots, endpoint=True)
    src_rows = np.linspace(0, rows, knots, endpoint=True)
    src_rows, src_cols = np.meshgrid(src_rows, src_cols)
    src = np.dstack([src_cols.flat, src_rows.flat])[0]

    # add random perturbation
    dst_rows = src[:, 1] + (np.random.randn(src.shape[0]) - 0.5) * intensity
    dst_cols = src[:, 0] + (np.random.randn(src.shape[0]) - 0.5) * intensity
    dst = np.vstack([dst_cols, dst_rows]).T

    # calculate average displacement
    d = np.mean(np.sqrt((dst - src)**2))

    tform = PiecewiseAffineTransform()
    tform.estimate(src, dst)

    out = warp(img, tform, output_shape=img.shape)

    return out, dst, d


def randomize_symbols(symbols, knots, intensity):
    symbols_ = []
    for S in symbols:
        S_, _, _ = random_nonrigid_transform(S, knots, intensity)
        symbols_.append(S_)
    return symbols_


def tilt3D(phi_x, phi_y, phi_z, method='interpolate'):
    # convert height into 3D points
    X = np.array(np.arange(self.size))
    points_3d = np.array([[float(X[i]), float(X[j]), self.die[X[i], X[j]]] for i in range(self.size) for j in range(self.size)])

    # rotation matrix in 3D space
    Rx = np.array([[1, 0, 0],
                   [0, np.cos(phi_x), -np.sin(phi_x)],
                   [0, np.sin(phi_x), np.cos(phi_x)]])
    Ry = np.array([[np.cos(phi_y), 0, np.sin(phi_y)],
                   [0, 1, 0],
                   [-np.sin(phi_y), 0, np.cos(phi_y)],
                   ])
    Rz = np.array([[np.cos(phi_z), -np.sin(phi_z), 0],
                   [np.sin(phi_z), np.cos(phi_z), 0],
                   [0, 0, 1]])
    R = Rx.dot(Ry).dot(Rz)

    # rotate and transform mask accordingly
    x, y, z = R.dot(points_3d.T)
    matrix = np.array([[np.cos(phi_x), 0, (1 - np.cos(phi_x)) * self.size // 2],
                       [0, np.cos(phi_y), (1 - np.cos(phi_y)) * self.size // 2],
                       [0, 0, 1]])
    tform = transform.AffineTransform(matrix=matrix)
    tilted_mask = transform.warp(self.mask, tform.inverse)

    # project onto Z²
    x = (x - x.min()).astype(int)
    y = (y - y.min()).astype(int)
    coords = np.array(list(zip(x, y)))

    tilted_coin = -np.ones((x.max() - x.min() + 1, y.max() - y.min() + 1))
    tilted_coin[coords[:, 0], coords[:, 1]] = z

    # smooth a little bit
    z = np.reshape(z, (1024, 1024))
    z = gaussian(z, 3)
    z = z.flatten()

    # complete the information lacking
    if method in ['interpolate']:
        grid_x, grid_y = np.mgrid[0: x.max() - x.min() + 1, 0: y.max() - y.min() + 1]
        tilted_coin = griddata(list(zip(x, y)), z, (grid_x, grid_y), method='cubic')
    elif method in ['inpainting']:
        holes = (255 * (tilted_coin == -1)).astype(np.uint8)
        tilted_coin = (255 * tilted_coin).astype(np.uint8)
        tilted_coin = cv.inpaint(tilted_coin, holes, 5, cv.INPAINT_TELEA)

    # normalize, smooth and project mask
    tilted_coin = (255 * normalize(tilted_coin)).astype(np.uint8)
    tilted_coin = gaussian(tilted_coin, 2)
    tilted_mask = tilted_mask[self.size // 2 - tilted_coin.shape[0] // 2: self.size // 2 + tilted_coin.shape[0] // 2, self.size // 2 - tilted_coin.shape[1] // 2: self.size // 2 + tilted_coin.shape[1] // 2]

    return tilted_coin, tilted_mask


class MetalPiece:
    def __init__(self, size,
                 radius=400,
                 thickness=500):
        self.size = size
        self.thickness = thickness
        self.radius = radius
        self.piece = np.zeros((size, size))
        self.mask = np.zeros((size, size))

    def resize(self, newsize):
        self.radius = self.radius * self.size / newsize
        self.piece = resize(self.piece, (newsize, newsize))
        self.mask = resize(self.mask, (newsize, newsize))
        self.size = newsize

    def take_a_picture(self, azimuth, slant, shot_noise=1, color=None):
        dx, dy = np.gradient(self.thickness * self.piece)
        dxNeg = dx * -1
        dyNeg = dy * -1

        dxSquare = np.power(dx, 2)
        dySquare = np.power(dy, 2)

        nxSquareRoot = np.sqrt(dxSquare + dxSquare + 1)
        nySquareRoot = np.sqrt(dySquare + dySquare + 1)
        nzSquareRoot = np.sqrt(dxSquare + dxSquare + 1)

        nx = np.divide(dxNeg, nxSquareRoot)
        ny = np.divide(dyNeg, nySquareRoot)
        nz = np.sqrt(1 - nx**2 - ny**2)

        n = np.array([nx, ny, nz]).T
        n = n / np.linalg.norm(n)
        l = np.array([np.cos(azimuth) * np.sin(slant), np.sin(azimuth) * np.sin(slant), np.cos(slant)])

        I = n.dot(l.T).T
        I = I + 0.0001 * np.random.poisson(shot_noise, size=(self.size, self.size))
        I = normalize(I) * self.mask

        if color is not None:
            I_rgb = np.zeros((self.size, self.size, 3), dtype=float)
            I_rgb[:, :, 0] = I * color[0]
            I_rgb[:, :, 1] = I * color[1]
            I_rgb[:, :, 2] = I * color[2]
            I = I_rgb

        return I


class Coin(MetalPiece):
    def __init__(self, size, radius=400, thickness=500):
        super().__init__(size, radius, thickness)
        self.texture = np.zeros((size, size))
        self.shift = np.array([size // 2, size // 2])

    def load_texture(self, texture):
        # randomize texture
        r = np.random.randn(texture.shape[0], texture.shape[1])
        texture += gaussian(r, 13)
        texture = rescale(texture, 1.5 * np.random.rand() + 0.5)
        texture = rotate(texture, 360 * np.random.rand())

        # standardize
        texture = resize(texture, (self.size, self.size))
        texture = gaussian(texture, 2)
        self.texture = normalize(texture)

    def generate_blank(self, smooth=1):
        blank = random_shape(self.size, self.radius, knots=11, amplitude=0.1)
        self.mask = blank
        blank = gaussian(blank, smooth)
        self.piece = normalize(blank)

    def strike(self, die, a=0.5, tilt=50, wearing_out=1):
        D = die.piece
        # shift center a little bit
        D_ = np.zeros(D.shape)
        dx = np.random.randint(2 * tilt) - tilt
        dy = np.random.randint(2 * tilt) - tilt
        self.shift = np.array([dx, dy])

        def non(s): return s if s < 0 else None
        def mom(s): return max(0, s)

        D_ = np.zeros_like(D)
        D_[mom(dy):non(dy), mom(dx):non(dx)] = D[mom(-dy):non(-dy), mom(-dx):non(-dx)]

        self.piece = self.piece - a * D_
        if wearing_out:
            die.piece = gaussian(die.piece, wearing_out)
        die.piece = normalize(die.piece)

    def wear_part(self, patch_size, x, y, strength=2):
        slice_ = np.s_[max(0, x - patch_size):min(x + patch_size, self.size),
                       max(0, y - patch_size): min(y + patch_size, self.size)]
        h = min(x + patch_size, self.size) - max(0, x - patch_size)
        w = min(y + patch_size, self.size) - max(0, y - patch_size)
        gmask = gaussian_window_mask((h, w), patch_size / 4)
        self.piece[slice_] = gaussian(self.piece[slice_], strength) * gmask + self.piece[slice_] * (1 - gmask)

    def wear(self, strength=1, part=False, radius=100, n_iter=1, texture_strength=0.1):
        if part:
            for i in range(n_iter):
                x, y = np.random.randint(self.size // 2 - self.radius, self.size // 2 + self.radius, 2)
                self.wear_part(radius, x, y, strength=strength)
        else:
            self.piece = gaussian(self.piece, 1)
            self.piece = normalize(self.piece - texture_strength * self.texture)


class Die(MetalPiece):
    def __init__(self, size, radius=100, thickness=500):
        super().__init__(size, radius, thickness)

        # create metal piece
        cx, cy = self.size // 2, self.size // 2
        die = np.zeros((self.size, self.size), dtype=float)

        x, y = np.meshgrid(np.arange(self.size), np.arange(self.size))
        x = x - cx
        y = y - cy

        # restrict metal piece to a disk and print writable area
        coords_mask = np.where(x**2 + y**2 < radius**2)
        die[coords_mask[0], coords_mask[1]] = 1.0
        knots = np.random.permutation([5, 7, 11, 19, 59, 89])[0]
        die = random_shape(size, radius, knots=knots, amplitude=0.05)

        self.piece = gaussian(die, 2)
        self.mask = die

    def carve_symbol(self, symbol, position, strength=0.5, rotation=0, scaling=1):
        symbol = rotate(symbol, rotation)
        symbol = rescale(symbol, scaling)

        xs, ys = position

        # apply the symbol to the die
        w, h = symbol.shape
        w = w - 1 * (w % 2 != 0)
        h = h - 1 * (h % 2 != 0)
        symbol = resize(symbol, (w, h))

        symbol_surface = np.zeros(self.piece.shape)
        blob_coords = np.argwhere(symbol)
        blob = symbol[blob_coords[:, 0], blob_coords[:, 1]]
        symbol_surface[blob_coords[:, 0] - xs - symbol.shape[0] // 2,
                       blob_coords[:, 1] - ys - symbol.shape[1] // 2] = blob
        symbol_surface = normalize(symbol_surface)
        self.piece -= strength * symbol_surface * self.mask

    def carve(self, symbols, positions, strength=0.5, rotation=0, scaling=1):
        for i in range(len(symbols)):
            S = symbols[i]
            pos = positions[i]
            self.carve_symbol(S, pos, strength=strength, rotation=rotation, scaling=scaling)
